/*
 * Ged.java - GUI-ed. A GUI front-end to the ed line editor.
 *            Written in Java for Unix/Linux platforms (needs /bin/ed).
 *
 * 09-May-2006, ver 0.70
 *
 * COPYRIGHT: Copyright (c) 2006 Brendan Gregg.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *  (http://www.gnu.org/copyleft/gpl.html)
 *
 * 09-May-2006  Brendan Gregg   Created this.
 */

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;        
import javax.swing.ImageIcon;        

/**
 * Ged - Gui-ED
 *
 * @author Brendan Gregg
 */
public class Ged implements KeyListener {

    private final int WIDTH = 80;           // default width, chars
    private final int HEIGHT = 16;          // default height, chars
    private LineNumberReader coprocInput;   // input from co-process
    private OutputStream coprocOutput;      // output to co-process
    private JFrame gedFrame;                // main frame
    private JTextField gedEntryWindow;      // text entry window
    private JTextArea gedOutputWindow;      // main output window
    private JCheckBoxMenuItem gedSoundCB;   // sound checkbox
    private SoundEffects sounds;            // sound effect player
    private boolean running;                // co-process running
    private boolean debug;                  // debug output
    private boolean soundOn;                // sound active
    private boolean soundLoaded;            // sounds loaded

    public enum OptionsSoundEvent { ON, OFF, TOGGLE }

    // constructor
    public Ged() {
        running = false;
        debug = false;
        soundOn = false;
        soundLoaded = false;
    }

    // debug method
    public void printDebug(String text) {
        if (debug)
            System.out.println(text);
    }

    public void exit(int exitStatus) {
        if (exitStatus >= 0) {
            System.out.println("Thank you for using Ged.");
            System.out.println("Arrrgh!! Command Line!");
        }
        System.exit(exitStatus);
    }

    /**
     * GedRead is a runnable class for a thread that reads output from
     * the co-process, and displays it on the main window.
     */
    private class GedRead implements Runnable {
        private byte b;
        private String s;

        public void run() {
            while (running) {
                // fetch output from co-process
                try {
                    s = coprocInput.readLine();
                    if (s == null)
                        break;
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }

                // print output from co-process
                printDebug("Output: " + s);
                gedOutputWindow.append(s + "\n");

                // make scrolling work
                gedOutputWindow.setCaretPosition(
                        gedOutputWindow.getDocument().getLength()); 
                try {
                    Thread.currentThread().sleep(10);
                } catch (InterruptedException exception) {
                }
            }
            running = false;

            // exit when ed exits
            exit(0);
        }
    }

    /**
     * runCommand runs the command specified, and creates a reader
     * thread to fetch its output.
     */
    public void runCommand(String cmd[]) {
        // don't start two
        if (running)
            return;

        // start command
        try {
            Process pid = new ProcessBuilder(cmd).start();
            running = true;
            coprocInput = new LineNumberReader(
                    new InputStreamReader(pid.getInputStream()));
            coprocOutput = pid.getOutputStream();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        // spawn reader
        Thread coprocReader = new Thread(new GedRead());
        coprocReader.start();
    }

    private void doFileOpen() {
        JFileChooser chooser = new JFileChooser();
        int returnVal = chooser.showOpenDialog(gedFrame);
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            String cmd = "e " + 
                chooser.getSelectedFile().getAbsolutePath() + "\n";
            try {
                coprocOutput.write(cmd.getBytes());
                coprocOutput.flush();
            } catch (IOException exception) {
                    gedOutputWindow.setText("ERROR: open failed\n");
            }
            printDebug("Running: " + cmd);
        }
    }

    private void doFileSave() {
        try {
            coprocOutput.write("w\n".getBytes());
            coprocOutput.flush();
        } catch (IOException exception) {
            gedOutputWindow.setText("ERROR: save failed\n");
        }
    }

    private void setSound(OptionsSoundEvent action) {
        if (! soundLoaded) {
            sounds = new SoundEffects(
                    "audio/click.wav", "audio/return.wav", "audio/clink.wav");
            soundLoaded = true;
        }
        if (action == OptionsSoundEvent.TOGGLE) {
            soundOn = gedSoundCB.getState();
        } else if (action == OptionsSoundEvent.ON) {
            soundOn = true;
            gedSoundCB.setState(true);
        } else if (action == OptionsSoundEvent.OFF) {
            soundOn = false;
            gedSoundCB.setState(false);
        }
    }

    private void setFontSize(int newSize) {
        if (newSize < 2 || newSize > 100) {
            newSize = 12;
        }
        gedEntryWindow.setFont(new Font("Courier New", Font.PLAIN, newSize));
        gedOutputWindow.setFont(new Font("Courier New", Font.BOLD, newSize));
        gedFrame.pack();
    }

    private void setForeground(Color newColor) {
        if (newColor != null) {
            gedOutputWindow.setForeground(newColor);
            gedEntryWindow.setForeground(newColor);
            printDebug("FG Color: " + newColor.getRed() + ", "
                    + newColor.getGreen() + ", " + newColor.getBlue());
        }
    }

    private void setBackground(Color newColor) {
        if (newColor != null) {
            gedOutputWindow.setBackground(newColor);
            gedEntryWindow.setBackground(newColor);
            printDebug("BG Color: " + newColor.getRed() + ", "
                    + newColor.getGreen() + ", " + newColor.getBlue());
        }
    }

    private void setColorTheme(String theme) {
        Color foreground;
        Color background;

        if (theme == "Classic") {
            foreground = new Color(0, 200, 0);
            background = new Color(0, 50, 50);
        } else if (theme == "Amber") {
            foreground = new Color(200, 200, 0);
            background = new Color(50, 50, 0);
        } else if (theme == "Old Ribbon") {
            foreground = new Color(100, 80, 100);
            background = new Color(240, 240, 230);
        } else if (theme == "Black") {
            foreground = new Color(245, 245, 245);
            background = new Color(0, 0, 0);
        } else {
            foreground = new Color(0, 0, 0);
            background = new Color(255, 255, 255);
        }

        setForeground(foreground);
        setBackground(background);
    }

    private void doVersion() {
        JOptionPane.showMessageDialog(gedFrame,
                "GUI-ed, 09-May-2006, ver 0.70", "Version",
                JOptionPane.INFORMATION_MESSAGE);
    }

    private void doGuide() {
        Runtime rt = Runtime.getRuntime();
        try {
            Process p = rt.exec("/usr/openwin/bin/xterm -e man ed &");
        } catch (java.io.IOException exception) {
            System.out.println("ERROR: on exec");
        }
    }

    private void doLicense() {
        JOptionPane.showMessageDialog(gedFrame,
            "COPYRIGHT: Copyright (c) 2006 Brendan Gregg.\n\n" +
            "This program is free software; you can redistribute it and/or\n" +
            "modify it under the terms of the GNU General Public License\n" +
            "as published by the Free Software Foundation; either version 2" +
            "\nof the License, or (at your option) any later version.\n\n" +
            "This program is distributed in the hope that it will be" +
            " useful,\nbut WITHOUT ANY WARRANTY; without even the implied" +
            " warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR" +
            " PURPOSE.  See the\nGNU General Public License for more" +
            " details.\n\nYou should have received a copy of the GNU" +
            " General Public License\nalong with this program; if not," +
            " write to the Free Software Foundation,\nInc., 59 Temple" +
            " Place - Suite 330, Boston, MA  02111-1307, USA.\n\n" +
            "(http://www.gnu.org/copyleft/gpl.html)\n", "License",
                JOptionPane.INFORMATION_MESSAGE);
    }

    /*
     * The following is an inner class (rather than an anonymous class) as
     * this definition is used several times.
     */
    private class GedMenuOptionsColorTheme implements ActionListener {
        public void actionPerformed(ActionEvent event) {
            JMenuItem source = (JMenuItem)(event.getSource());
            String theme = source.getText();
            setColorTheme(theme);
        }
    }

    // keyboard enter key - send command
    private class GedEntryWindowHandler implements ActionListener {
        public void actionPerformed(ActionEvent event) {
            printDebug("Entry: " + gedEntryWindow.getText());
            try {
                if (coprocOutput != null) {
                    coprocOutput.write(gedEntryWindow.getText().getBytes());
                    coprocOutput.write('\n');
                    coprocOutput.flush();
                }
            } catch (IOException e) {
                gedOutputWindow.append("ERROR: writing\n");
            }
            gedEntryWindow.setText("");
        }
    }

    // keyboard pressed - check for sound effects
    public void keyPressed(KeyEvent event) {
        char c = event.getKeyChar();
        printDebug("Key: " + c);
        if (soundOn)
        {
            if (c == '\n')
                sounds.play("audio/return.wav");    // ping-woosh
            else
                sounds.play("audio/click.wav");     // click

            if (gedEntryWindow.getText().length() == 72)
                sounds.play("audio/clink.wav");     // margin ping
        }
    }
    public void keyReleased(KeyEvent event) { }
    public void keyTyped(KeyEvent event) { }

    /**
     * createAndShowGUI creates the main window for Ged.
     */
    private void createAndShowGUI() {
        /*
         * Create Frame
         */
        JFrame.setDefaultLookAndFeelDecorated(true);
        gedFrame = new JFrame("Ged - GUI ed");
        gedFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        /*
         * Display Pointless Splash Screen
         */
        SplashScreen splash = new SplashScreen(gedFrame);
        splash.run();

        /*
         * Create Window Components
         */
        gedEntryWindow = new JTextField(WIDTH);
        gedEntryWindow.addKeyListener(this);
        gedEntryWindow.setFont(new Font("Courier New", Font.PLAIN, 16));
        gedOutputWindow = new JTextArea(HEIGHT, WIDTH);
        gedOutputWindow.setFont(new Font("Courier New", Font.BOLD, 16));
        gedOutputWindow.setEditable(false);
        JScrollPane scroller = new JScrollPane(gedOutputWindow);
        scroller.setVerticalScrollBarPolicy(
                JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        setColorTheme("Black");

        /*
         * Create File Menu
         */
        JMenuBar gedMenuBar = new JMenuBar();
        JMenu gedMenu = new JMenu("File");
        gedMenuBar.add(gedMenu);
        // menu -> file -> open,
        JMenuItem menuItem = new JMenuItem("Open");
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                doFileOpen();
            }
        });
        gedMenu.add(menuItem);
        // menu -> file -> save,
        menuItem = new JMenuItem("Save");
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                doFileSave();
            }
        });
        gedMenu.add(menuItem);
        gedMenu.addSeparator();
        // menu -> file -> exit,
        menuItem = new JMenuItem("Exit");
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                exit(0);
            }
        });
        gedMenu.add(menuItem);

        /*
         * Create Options Menu
         */
        gedMenu = new JMenu("Options");
        gedMenuBar.add(gedMenu);
        // menu -> options -> sound,
        gedSoundCB = new JCheckBoxMenuItem("Sound");
        gedSoundCB.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                setSound(OptionsSoundEvent.TOGGLE);
            }
        });
        gedMenu.add(gedSoundCB);
        gedMenu.addSeparator();
        // menu -> options -> font size,
        JMenu subMenu = new JMenu("Font Size");
        for (int i = 6; i < 24; i += 2) {
            menuItem = new JMenuItem("" + i);
            subMenu.add(menuItem);
            menuItem.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    int size;
                    JMenuItem source = (JMenuItem)(event.getSource());
                    try {
                        size = Integer.parseInt(source.getText());
                    } catch (NumberFormatException e) {
                        size = 16;
                    }
                    setFontSize(size);
                }
            });
        }
        gedMenu.add(subMenu);
        gedMenu.addSeparator();
        // menu -> options -> foreground color,
        menuItem = new JMenuItem("Foreground Color");
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                Color newColor = JColorChooser.showDialog(gedFrame, 
                        "Choose Foreground Color",
                        gedOutputWindow.getForeground());
                setForeground(newColor);
            }
        });
        gedMenu.add(menuItem);
        // menu -> options -> background color,
        menuItem = new JMenuItem("Background Color");
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                Color newColor = JColorChooser.showDialog(gedFrame,
                        "Choose Background Color",
                        gedOutputWindow.getBackground());
                setBackground(newColor);
            }
        });
        gedMenu.add(menuItem);
        // menu -> options -> color theme,
        subMenu = new JMenu("Color Theme");
        menuItem = new JMenuItem("Black");
        menuItem.addActionListener(new GedMenuOptionsColorTheme());
        subMenu.add(menuItem);
        menuItem = new JMenuItem("White");
        menuItem.addActionListener(new GedMenuOptionsColorTheme());
        subMenu.add(menuItem);
        menuItem = new JMenuItem("Classic");
        menuItem.addActionListener(new GedMenuOptionsColorTheme());
        subMenu.add(menuItem);
        menuItem = new JMenuItem("Amber");
        menuItem.addActionListener(new GedMenuOptionsColorTheme());
        subMenu.add(menuItem);
        menuItem = new JMenuItem("Old Ribbon");
        menuItem.addActionListener(new GedMenuOptionsColorTheme());
        subMenu.add(menuItem);
        gedMenu.add(subMenu);

        /*
         * Create Help Menu
         */
        gedMenu = new JMenu("Help");
        gedMenuBar.add(Box.createHorizontalGlue());
        gedMenuBar.add(gedMenu);
        // menu -> help -> version,
        menuItem = new JMenuItem("Version");
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                doVersion();
            }
        });
        gedMenu.add(menuItem);
        // menu -> help -> guide,
        menuItem = new JMenuItem("Guide");
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                doGuide();
            }
        });
        gedMenu.add(menuItem);
        // menu -> help -> license,
        menuItem = new JMenuItem("License");
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                doLicense();
            }
        });
        gedMenu.add(menuItem);
        gedFrame.setJMenuBar(gedMenuBar);

        // create layout
	    GridBagLayout grid = new GridBagLayout();
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.fill = GridBagConstraints.BOTH;
        gbc.weightx = 1;
        gbc.weighty = 1;
        grid.setConstraints(scroller, gbc);
        gbc.weighty = 0;
        gbc.gridy = 1;
        grid.setConstraints(gedEntryWindow, gbc);
        JPanel gedPanel = new JPanel(grid);
        gedPanel.add(scroller);
        gedPanel.add(gedEntryWindow);
        gedFrame.getContentPane().add(gedPanel, BorderLayout.CENTER);

        // add actions
        gedEntryWindow.addActionListener(new GedEntryWindowHandler());

        // display the window.
        gedFrame.pack();
        gedFrame.setVisible(true);
        gedEntryWindow.requestFocusInWindow();
    }

    /**
     * main - Ged main program,
     * display gui and run command.
     */
    public static void main(String args[]) {
        Ged ged = new Ged();
        String[] cmd = new String[2];

        // check and set args
        if (args.length > 1) {
            System.err.println("USAGE: ged [filename]");
            ged.exit(-1);
        }
        cmd[0] = "/usr/bin/ed";
        if (args.length == 1) {
            cmd[1] = args[0];
        } else {
            cmd[1] = "";
        }

        // run ed
        ged.createAndShowGUI();
        ged.runCommand(cmd);
    }
}
