/*
 * SoundEffects - a sound effect engine.
 *
 * 09-May-2006, ver 0.50
 *
 * COPYRIGHT: Copyright (c) 2006 Brendan Gregg.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *  (http://www.gnu.org/copyleft/gpl.html)
 *
 * 09-May-2006  Brendan Gregg   Created this.
 */

import java.util.*;
import javax.sound.sampled.*;

/**
 * A simple 4-channel mixer to play short sound effects.
 * These sound effects are cached in RAM, by the SoundClip object.
 */
public class SoundEffects {

    // variables
    private Hashtable<String, SoundClip> soundHash;
    private SourceDataLine[] oldLines;
    private int maxIndex;
    private SoundClip defaultClip;

    /**
     * The constructor takes a list of sound filenames and preloads them.
     * At least 1 filename must be provided. The first filename will be
     * used as the default sound if an invalid filename is used in the
     * play() method.
     */
    public SoundEffects(String fileName, String ... fileNames) {
        int bytes;
        maxIndex = fileNames.length;            // yes, really
        oldLines = new SourceDataLine[3];       // concurrent sounds
        soundHash = new Hashtable<String, SoundClip>();

        // load files
        defaultClip = new SoundClip(fileName);
        soundHash.put(fileName, defaultClip);
        for (int index = 0; index < maxIndex; index++) {
            soundHash.put(fileNames[index], new SoundClip(fileNames[index]));
        }

        // populate previous output line
        try { 
            oldLines[2] = (SourceDataLine)
                    AudioSystem.getLine(defaultClip.info);
        } catch (LineUnavailableException e) {
            System.out.println("ERROR (/dev/audio): " + e.toString());
            System.exit(-1);
        }
    }

    /**
     * The play method takes a filename and plays the sound from its cache.
     * This creates a new sound line, so that we can play sounds concurrently.
     * If the filename is invalid, a default sound is choosen.
     */
    public void play(String fileName) {
        SoundClip clip;

        // choose sample
        if (soundHash.containsKey(fileName)) {
            clip = (SoundClip) soundHash.get(fileName);
        } else {
            clip = defaultClip;
        }

        // queue sample to be played on a new line
        try {
            // fetch new output line
            SourceDataLine line = (SourceDataLine)
                    AudioSystem.getLine(clip.info);

            // begin playback
            line.open(clip.format);
            line.start();
            int bytes = line.write(clip.sample, 0, clip.bytes);

            // close previous playing line
            if (oldLines[2] != null)
                oldLines[2].close();
            oldLines[2] = oldLines[1];
            oldLines[1] = oldLines[0];
            oldLines[0] = line;

        } catch (LineUnavailableException e) {
            System.out.println("ERROR (new line): " + e.toString());
        }
    }
}
