/*
 * SoundLoop.java - play a sound on a loop.
 *
 * 09-May-2006, ver 0.50
 *
 * COPYRIGHT: Copyright (c) 2006 Brendan Gregg.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *  (http://www.gnu.org/copyleft/gpl.html)
 *
 * 09-May-2006  Brendan Gregg   Created this.
 */

import java.io.File;
import java.lang.Class;
import java.net.URL;
import javax.sound.sampled.*;

/**
 * Play a sound on a loop.
 * This creates a thread that will continually play the sound file
 * provided in the constructor. This may be suitable for background
 * music.
 */
public class SoundLoop {
    private static final int BUFSIZE = 131072;
    private String fileName;
    private URL url;
    private SourceDataLine line;
    private boolean running;

    /**
     * Create a SoundLoop object from the provided sound file.
     * This currently only accepts .wav files.
     */
    public SoundLoop(String fileName) {
        this.fileName = fileName;
        url = getClass().getResource(fileName);
        if (url == null) {
            System.err.println("ERROR: SoundLoop couldn't open " + fileName);
        }
        running = false;
    }

    public void stop() {
        running = false;
        line.close();
    }
        
    public void play() {
        Thread thread = new Thread(new SoundLoopPlayer());
        thread.start();
    }

    private class SoundLoopPlayer implements Runnable {
        int bytesIn;
        int bytesOut;
        byte[] data;

        public void run() {
            try {
                // open audio file
                AudioInputStream audio = AudioSystem.getAudioInputStream(url);
                AudioFormat format = audio.getFormat();
                DataLine.Info info = 
                        new DataLine.Info(SourceDataLine.class, format);

                // open line to audio device
                line = (SourceDataLine) AudioSystem.getLine(info);
                line.open(format);
                line.start();
                running = true;

                // play in loop
                while (running) {
                    bytesIn = 0;
                    data = new byte[BUFSIZE];
                    while (bytesIn != -1) {
                        bytesIn = audio.read(data, 0, data.length);
                        if (bytesIn >= 0) {
                            bytesOut = line.write(data, 0, bytesIn);
                        }
                    }
                    audio = AudioSystem.getAudioInputStream(url);
                }

                // close audio line
                line.drain();
                line.close();

            } catch (Exception e) {
                e.printStackTrace();
                System.exit(1);
            }
        }
    }
}
